/*
 *
 *  MikMOD driver for the Advanded Linux Sound Architecture - sequencer
 *
 */

#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/asoundlib.h>
#include <linux/ainstr_simple.h>

#include "mikmod.h"

#define MAX_INSTRUMENTS		128	/* Max. instruments loadable   */
#define MAX_CHANNELS		128	/* Max. channels available */

#define	CH_FREQ		1
#define CH_VOL		2
#define	CH_PAN		4

extern int seq_dest_client;
extern int seq_dest_port;
extern snd_seq_t *seq_handle;

typedef unsigned short word;

struct actualPosition {
	word order;
	word row;
};

struct SAMPLE_DATA {	/* This structure is used to hold information  */
  char *sample;		/* about a loaded sample. The sample is also   */
  ULONG length;		/* kept in normal memory. Usefull for spectrum */
  ULONG loopstart;	/* or whatever analyzing of the music (later)  */
  ULONG loopend;
  UWORD flags;
  UWORD active;
};


struct VOICE {		/* A structure to hold the current state	*/
  UBYTE in_use;		/* of a GUS voice channel. Again usefull for	*/
  int handle;		/* music analyzing				*/
  UWORD flags;
  ULONG start;
  ULONG size;
  ULONG reppos;
  ULONG repend;
  ULONG freq;
  int kick;
  int changes;
  short vol;
  short pan;
};

/* Global declarations follow */

struct SAMPLE_DATA instrs[MAX_INSTRUMENTS];

struct VOICE voices[MAX_CHANNELS];

int nr_voices, nr_instrs;
int oldRow=0, oldOrder=0;
int seq_fd;
int seq_queue;
int seq_client;
int seq_port;
int seq_timer = 0;
snd_seq_tick_time_t seq_time = 0;
static UBYTE ALSA_BPM;


static BOOL AlsaIsThere(void)
{
	return 1;
}


static void BuildEvent(snd_seq_event_t *ev, int type)
{
	fprintf(stderr, "BuildEvent - %i\n", type);
	bzero(ev, sizeof(*ev));
	ev->time.tick = seq_time;
	ev->source.queue = seq_queue;
	ev->source.client = seq_client;
	ev->source.port = seq_port;
	ev->dest.queue = seq_queue;
	ev->dest.client = seq_dest_client;
	ev->dest.port = seq_dest_port;
	ev->type = type;
}

static BOOL ResetCard(void) {
	snd_seq_event_t ev;
	snd_seq_port_subscribe_t sub;
	int err;

#if 0
	if ( ultra_reset( ultra_handle, nr_voices, 0) < 0 ) {
		return 0;
	}
#endif
	if (seq_timer) {
		BuildEvent(&ev, SND_SEQ_EVENT_STOP);
		ev.dest.client = SND_SEQ_CLIENT_SYSTEM;
		ev.dest.port = SND_SEQ_PORT_SYSTEM_TIMER;
		if (snd_seq_event_output(seq_handle, &ev)<0)
			fprintf(stderr, "Unable to write a TIMER STOP event.\n");
		while (snd_seq_flush_output(seq_handle)>0)
			sleep(1);
		seq_timer = 0;
	}
	seq_time = 0;

	bzero(&sub, sizeof(sub));
	sub.sender.queue = seq_queue;
	sub.sender.client = seq_client;
	sub.sender.port = seq_port;
	sub.dest.queue = seq_queue;
	sub.dest.client = seq_dest_client;
	sub.dest.port = seq_dest_port;
	
	if ((err = snd_seq_unsubscribe_port(seq_handle, &sub))<0) {
		snd_seq_close(seq_handle);
		seq_handle = NULL;
		fprintf(stderr, "Unable to unsubcribe destonation port: %s\n", snd_strerror(err));
		exit(0);
	}

	sub.exclusive = 1;
	sub.synth_voices = nr_voices;

	if ((err = snd_seq_subscribe_port(seq_handle, &sub))<0) {
		snd_seq_close(seq_handle);
		seq_handle = NULL;
		fprintf(stderr, "Unable to subcribe destonation port: %s\n", snd_strerror(err));
		exit(0);
	}

        return 1;
}


static BOOL AlsaInit(void)
{
	snd_seq_client_info_t cinfo;
	snd_seq_port_info_t port;
	snd_seq_port_subscribe_t sub;
	int err;

	if (!(md_mode & DMODE_16BITS)) {
		md_mode |= DMODE_16BITS;	/* GUS doesn't do 8-bit */
	}
	if (!(md_mode & DMODE_STEREO)) {	/* and also no MONO */
		md_mode |= DMODE_STEREO;
	}

	md_mixfreq = 44100;		/* and also ignore freq. changes */

	if ((err = snd_seq_open(&seq_handle, SND_SEQ_OPEN))<0) {
		myerr = (char *)snd_strerror(err);
		return 0;
	}

	seq_client = snd_seq_client_id(seq_handle);
	if (seq_client < 0) {
		snd_seq_close(seq_handle);
		seq_handle = NULL;
		myerr = "Unable to determine my client number...";
		return 0;
	}

	bzero(&cinfo, sizeof(cinfo));
	cinfo.client = seq_client;
	cinfo.type = USER_CLIENT;
	snprintf(cinfo.name, sizeof(cinfo.name), "alsamod - %i", getpid());
	if ((err = snd_seq_set_client_info(seq_handle, &cinfo)) < 0) {
		snd_seq_close(seq_handle);
		seq_handle = NULL;
		myerr = "Unable to set client info...";
		return 0;
	}

	seq_queue = snd_seq_alloc_queue(seq_handle);
	if (seq_queue < 0) {
		snd_seq_close(seq_handle);
		seq_handle = NULL;
		myerr = "Unable to allocate a new queue...";
		return 0;
	}

	if ((err = snd_seq_block_mode(seq_handle, 0))<0) {
		snd_seq_close(seq_handle);
		seq_handle = NULL;
		myerr = "Unable to set the non-block mode...";
		return 0;
	}
	
	bzero(&port, sizeof(port));
	strcpy(port.name, "Output");
	if ((err = snd_seq_create_port(seq_handle, &port)) < 0) {
		snd_seq_close(seq_handle);
		seq_handle = NULL;
		myerr = "Unable to create a client port...";
		return 0;
	}
	seq_port = port.port;

	bzero(&sub, sizeof(sub));
	sub.sender.queue = seq_queue;
	sub.sender.client = seq_client;
	sub.sender.port = seq_port;
	sub.dest.queue = seq_queue;
	sub.dest.client = seq_dest_client;
	sub.dest.port = seq_dest_port;
	sub.exclusive = 1;
	sub.synth_voices = nr_voices = 4;
	
	if ((err = snd_seq_subscribe_port(seq_handle, &sub))<0) {
		snd_seq_close(seq_handle);
		seq_handle = NULL;
		myerr = "Unable to subcribe destonation port...";
		return 0;
	}

	if ((err = snd_seq_get_any_port_info(seq_handle, seq_dest_client, seq_dest_port, &port))<0) {
		snd_seq_close(seq_handle);
		seq_handle = NULL;
		myerr = "Unable to get destonation port...";
		return 0;
	}
	
	puts(port.name);
	puts("\n");

	seq_fd = snd_seq_file_descriptor(seq_handle);

	return 1;
}


static unsigned int getSampleType( struct SAMPLE_DATA *smp) {
	unsigned int result = 0;

	if (smp->flags & SF_DELTA) {
		fprintf(stderr, "DELTA SAMPLES ARE NOT SUPPORTED YET");
		exit(0);
	}
	if (smp->flags & SF_LOOP)
		printf("LOOOPPPPPPPPPPPPPPPPPPPPP\n");
	result = ((smp->flags & SF_16BITS) ? SIMPLE_WAVE_16BIT : 0)  |
		 ((smp->flags & SF_LOOP) ? SIMPLE_WAVE_LOOP : 0) |
		 ((smp->flags & SF_BIDI) ? SIMPLE_WAVE_BIDIR : 0);
	return result;
}


int LoadSamples() {
	struct SAMPLE_DATA *smp;
	simple_instrument_t instrument;
	snd_seq_instr_put_t *put;
	ULONG length, loopstart, loopend;
	int i, err;
	long size;
	char name[80];
	snd_seq_event_t ev, *pev;
	fd_set in;
	struct timeval tv;
		
	for (i = 0; i < nr_instrs; i++) {
		smp = &instrs[i];
		
		if (!(smp -> active)) {
			continue;
		}
		length = smp->length;
		loopstart = smp->loopstart;
		loopend = smp->loopend;

		if (smp->flags & SF_16BITS) {
			length<<=1;
			loopstart<<=1;
			loopend<<=1;
		}

		memset(&instrument, 0, sizeof(instrument));
		instrument.format	= getSampleType(smp);
		printf("instrument.format = 0x%x\n", instrument.format);
		instrument.address.ptr	= smp->sample;
		instrument.loop_start	= loopstart << 4;
		instrument.loop_end	= loopend << 4;
		instrument.size		= length;
		if (smp->flags & SF_LOOP) {
			smp->sample[loopend] = smp->sample[loopstart];
			if (smp->flags & SF_16BITS) {
				smp->sample[loopend-1] = smp->sample[loopstart-1];
			}
		}
		
		sprintf(name, "AlsaMod Wave %i", i);
		if (snd_instr_simple_convert_to_stream(&instrument, name, &put, &size)<0) {
			fprintf(stderr, "Unable to convert a sample to stream...\n");
			continue;
		}
		
		put->id.std = 1 << 24;	/* private instrument */
		put->id.prg = i;		

		BuildEvent(&ev, SND_SEQ_EVENT_INSTR_PUT);
		ev.flags |= SND_SEQ_EVENT_LENGTH_VARUSR | SND_SEQ_DEST_DIRECT;
		ev.data.ext.len = size;
		ev.data.ext.ptr = put;

		if (snd_seq_event_output(seq_handle, &ev)<0) {
			fprintf(stderr, "Unable to write a SAMPLE PUT event...\n");
			continue;
		}
		/* ugly, but working */
		while (snd_seq_flush_output(seq_handle)>0)
			sleep(1);
		
		while (1) {
			FD_ZERO(&in);
			FD_SET(seq_fd, &in);
			tv.tv_sec = 30;
			tv.tv_usec = 0;
			if ((err = select(seq_fd + 1, &in, NULL, NULL, &tv)) < 0) {
				fprintf(stderr, "Sample select error...\n");
				goto __next;
			}
			if (err == 0) {	/* timeout */
				fprintf(stderr, "Sample download timeout...\n");
				goto __next;
			}
			if (FD_ISSET(seq_fd, &in)) {
				do {
					if ((err = snd_seq_event_input(seq_handle, &pev)) < 0)
						break;
					if (!pev)
						continue;
					fprintf(stderr, "PUT: pev->type = %i\n", pev->type);
					if (pev->type == SND_SEQ_EVENT_RESULT) {
						if (pev->data.result.event == SND_SEQ_EVENT_INSTR_PUT) {
							fprintf(stderr, "PUT: result = %i\n", pev->data.result.result);
							if (pev->data.result.result < 0) {
								fprintf(stderr, "Sample download ERROR!!!\n");
								snd_seq_free_event(pev);
								goto __next;
							}
							snd_seq_free_event(pev);
							goto __next;
						}
					}
				} while (err > 0);
			}
		}
		
	      __next:
	}
	return 0;
}


SWORD AlsaSampleLoad(FILE *fp,ULONG length,ULONG loopstart,ULONG loopend,UWORD flags)
{
	struct SAMPLE_DATA *smp;
	char *p;
	long l;

	smp = &instrs[nr_instrs];
	smp->length = length;
	smp->loopstart = loopstart;
	if ( !loopend )
	  smp->loopend = length - (flags & SF_16BITS ? 2 : 1);
	 else
	  smp->loopend = loopend;
	smp->flags = flags; 
	if ( ( smp->flags & SF_16BITS) ) {
                length<<=1;
                loopstart<<=1;
                loopend<<=1;
        }
	if ( ( smp->sample = (char *) malloc( length + 16 ) ) == NULL ) {
	   	myerr="not enough memory to load sample.";
		return -1;
	}	

	SL_Init(fp,flags,flags|SF_SIGNED);

	l = length;
	p = smp->sample;
	
	while( l > 0 ) {
	  /*		static UBYTE buffer[8192];*/
	        long todo;
        	todo=(l>8192) ? 8192 : l;
	        SL_Load(p ,todo);
                p+=todo;
                l-=todo;
	}	

	smp->active = 1;	
	
	return nr_instrs++;
}

static void AlsaSampleUnload(SWORD h)
{
	struct SAMPLE_DATA *smp;
	snd_seq_instr_free_t ifree;
	snd_seq_event_t ev, *pev;
	fd_set in;
	struct timeval tv;
	int err;

	if (h > MAX_INSTRUMENTS) {
		return;
	}	
	smp = &instrs[h];
		
	if (!(smp->active))
		return;

	bzero(&ifree, sizeof(ifree));
	ifree.cmd = SND_SEQ_INSTR_FREE_CMD_SINGLE;
	ifree.data.instr.std = 1 << 24;	/* private instrument */
	ifree.data.instr.prg = h;	
	BuildEvent(&ev, SND_SEQ_EVENT_INSTR_FREE);
	ev.flags |= SND_SEQ_EVENT_LENGTH_VARUSR | SND_SEQ_DEST_DIRECT;
	ev.data.ext.len = sizeof(ifree);
	ev.data.ext.ptr = &ifree;

	if (snd_seq_event_output(seq_handle, &ev)<0)
		fprintf(stderr, "Unable to write a SAMPLE FREE event...");

	/* ugly, but working */
	while (snd_seq_flush_output(seq_handle)>0)
		sleep(1);

	while (1) {
		FD_ZERO(&in);
		FD_SET(seq_fd, &in);
		tv.tv_sec = 30;
		tv.tv_usec = 0;
		if ((err = select(seq_fd + 1, &in, NULL, NULL, &tv)) < 0) {
			fprintf(stderr, "Sample free select error...");
			goto __end;
		}
		if (err == 0) {	/* timeout */
			fprintf(stderr, "Sample free timeout...");
			goto __end;
		}
		if (FD_ISSET(seq_fd, &in)) {
			do {
				if ((err = snd_seq_event_input(seq_handle, &pev)) < 0)
					break;
				if (!pev)
					continue;
				if (pev->type == SND_SEQ_EVENT_RESULT) {
					if (pev->data.result.event == SND_SEQ_EVENT_INSTR_FREE) {
						if (pev->data.result.result < 0) {
							fprintf(stderr, "Sample free ERROR!!!");
							snd_seq_free_event(pev);
							goto __end;
						}
						snd_seq_free_event(pev);
						goto __end;
					}
				}
			} while (err > 0);
		}
	}
		
     __end:	

	free(smp->sample);
	nr_instrs--;
	smp->active = 0;
}


static void AlsaPlayer(void)
{
	int t;
	struct VOICE *voice;
	snd_seq_event_t ev;

	md_player();

	for ( t = 0; t < md_numchn; t++ ) {
		voice = &voices[t];
#if 0
		if (t != 2)
			continue;
#endif
		if (voice->kick) {
			printf("%i: voice->kick\n", t);
			BuildEvent(&ev, SND_SEQ_EVENT_SAMPLE_STOP);
			ev.dest.channel = t;
			ev.data.stop_mode = SAMPLE_STOP_IMMEDIATELY;
			if (snd_seq_event_output(seq_handle, &ev)<0)
				fprintf(stderr, "Unable to write a SAMPLE STOP event...");
			BuildEvent(&ev, SND_SEQ_EVENT_SAMPLE);
			ev.dest.channel = t;
			ev.data.sample.bank = 0;
			ev.data.sample.prg = voice->handle;
			if (snd_seq_event_output(seq_handle, &ev)<0)
				fprintf(stderr, "Unable to write a SAMPLE SELECT event...");
			voice->changes |= CH_FREQ | CH_VOL | CH_PAN;
		}
		if (voice->changes & CH_FREQ) {
			BuildEvent(&ev, SND_SEQ_EVENT_SAMPLE_FREQ);
			ev.dest.channel = t;
			ev.data.frequency = voice->freq << 4;
			if (snd_seq_event_output(seq_handle, &ev)<0)
				fprintf(stderr, "Unable to write a SAMPLE FREQ event...");
		}
		if (voice->changes & (CH_VOL|CH_PAN)) {
			BuildEvent(&ev, SND_SEQ_EVENT_SAMPLE_VOLUME);
			ev.dest.channel = t;
			ev.data.volume.volume = (voice->vol << 8) - 1;
			if (ev.data.volume.volume < 0)
				ev.data.volume.volume = 0;
			ev.data.volume.lr = (voice->pan << 6) - 1;
			if (ev.data.volume.lr < 0)
				ev.data.volume.lr = 0;
			ev.data.volume.fr = 8192;
			ev.data.volume.du = 8192;
			if (snd_seq_event_output(seq_handle, &ev)<0)
				fprintf(stderr, "Unable to write a SAMPLE VOLUME event...");
		}
		voice->changes = 0;
		if (voice->kick) {
			voice->kick = 0;
			BuildEvent(&ev, SND_SEQ_EVENT_SAMPLE_START);
			ev.dest.channel = t;
			ev.data.position = voice->start;
			if (snd_seq_event_output(seq_handle, &ev)<0)
				fprintf(stderr, "Unable to write a SAMPLE START event...");
		}			
	}

	seq_time++;
}


#if 0
static void ProcessEcho( void *handle, void *private, unsigned char *data ) {
        struct actualPosition pos;
        
        switch ( *data ) { 
              case 0x00: memcpy( &pos, data+1, sizeof(pos));
                         printf("\rPattern/Pos: %03i/%03i          ",
                                 pos.order, pos.row);
                         fflush(stdout);
        }
}
#endif


#if 0
static void ProcessInput(void) {
	static ultra_callbacks_t callbacks = {
		2,			/* I know about 2 callbacks only... */
		NULL,			/* private */
		ProcessEcho,		/* call_echo */
		NULL			/* call_voices_change */
	};
	
	ultra_do_input( ultra_handle, &callbacks );
}
#endif
                                                                        
static void doEcho() {
#if 0
	struct actualPosition pos;
	unsigned char args[ 1 + sizeof( pos ) ];
	pos.order = mp_sngpos;
	pos.row = mp_patpos;
	
	args[ 0 ] = 0x00;
	memcpy(args+1, (void *)&pos, sizeof(pos));
	ultra_do_echo(ultra_handle, args, sizeof(args));
#endif
}



static void AlsaUpdate(void) {
	fd_set read_fds, write_fds;
	int need_write, err;
	snd_seq_event_t ev, *pev;

	if (ALSA_BPM != md_bpm) {
		BuildEvent(&ev, SND_SEQ_EVENT_TEMPO);
		ev.dest.client = SND_SEQ_CLIENT_SYSTEM;
		ev.dest.port = SND_SEQ_PORT_SYSTEM_TIMER;
		ev.data.control.value = md_bpm * 600000;
		if (snd_seq_event_output(seq_handle, &ev)<0)
			fprintf(stderr, "Unable to write a TIMER START event.\n");
#if 0
		ultra_do_tempo( ultra_handle, ((md_bpm * 50) / 125) * 1);
#endif
		
		ALSA_BPM = md_bpm;
	}

	AlsaPlayer();
	if (oldRow != mp_sngpos || oldOrder != mp_patpos) {
		doEcho();
		oldRow = mp_sngpos;
		oldOrder = mp_patpos;
	}
	do {
		need_write = snd_seq_flush_output(seq_handle);
		
		FD_ZERO(&read_fds);
                FD_ZERO(&write_fds);
		
		do {
			FD_SET(seq_fd, &read_fds);		
			FD_SET(seq_fd, &write_fds);

			select(seq_fd+1, &read_fds, &write_fds, NULL, NULL);
	
			if (FD_ISSET(seq_fd, &read_fds)) {
				do {
					if ((err = snd_seq_event_input(seq_handle, &pev)) < 0)
						break;
					if (!pev)
						continue;
					snd_seq_free_event(pev);
				} while (err > 0);
			}
		} while (!FD_ISSET(seq_fd, &write_fds));
	} while (need_write > 0);
}


static void AlsaExit(void)
{
	snd_seq_close(seq_handle);
	seq_handle = NULL;
	seq_fd = -1;
}


static void AlsaPlayStart(void) {
	int t;
	snd_seq_event_t ev;

	for ( t = 0; t < MAX_CHANNELS; t++ ) {
		voices[ t ].flags = 0;
		voices[ t ].handle = 0;
		voices[ t ].size = 0;
		voices[ t ].start = 0;
		voices[ t ].reppos = 0;
		voices[ t ].repend = 0;
		voices[ t ].changes = 0;
		voices[ t ].kick = 0;
		voices[ t ].freq = 10000;
		voices[ t ].vol = 64;
		voices[ t ].pan = 128;
	}
	
	nr_voices = md_numchn;
	

	if ( ResetCard() == 0 ) {
		puts("ALSA: problem resetting card!");
		exit(-1);
	}
	LoadSamples();

#if 0	
	ultra_queue_write_set_size( ultra_handle, 1024 );
	ultra_queue_read_set_size( ultra_handle, 128 );
	
	if ( ultra_timer_start( ultra_handle ) < 0 ) {
		puts("ALSA: Error starting timer");
	}
	ultra_timer_tempo(ultra_handle, 50);
	ULTRA_BPM = 0;
#endif

	BuildEvent(&ev, SND_SEQ_EVENT_TEMPO);
	ev.dest.client = SND_SEQ_CLIENT_SYSTEM;
	ev.dest.port = SND_SEQ_PORT_SYSTEM_TIMER;
	ev.data.control.value = 500000;
	if (snd_seq_event_output(seq_handle, &ev)<0)
		fprintf(stderr, "Unable to write a TIMER TEMPO event.\n");

	BuildEvent(&ev, SND_SEQ_EVENT_START);
	ev.dest.client = SND_SEQ_CLIENT_SYSTEM;
	ev.dest.port = SND_SEQ_PORT_SYSTEM_TIMER;
	if (snd_seq_event_output(seq_handle, &ev)<0)
		fprintf(stderr, "Unable to write a TIMER START event.\n");

	while (snd_seq_flush_output(seq_handle)>0)
		sleep(1);
}


static void AlsaPlayStop(void)
{
#if 0
	ultra_queue_flush( ultra_handle );
	ultra_timer_stop( ultra_handle );

	/* ok, there maybe some waiting echo data - Perex */

	for ( i = 0; i < 10; i++ )
	  ProcessInput();

	ultra_queue_write_set_size( ultra_handle, 0 );
	ultra_queue_read_set_size( ultra_handle, 0 );
#endif
}


static void AlsaVoiceSetVolume(UBYTE voice,UBYTE vol)
{
	if (vol != voices[voice].vol) {
		voices[voice].vol = vol;
		voices[voice].changes |= CH_VOL;
	}
}


static void AlsaVoiceSetFrequency(UBYTE voice,ULONG freq)
{
	if (freq != voices[voice].freq) {
		voices[voice].freq = freq;
		voices[voice].changes |= CH_FREQ;
	}
}


static void AlsaVoiceSetPanning(UBYTE voice,UBYTE pan)
{
	if (pan != voices[voice].pan) {
		voices[voice].pan = pan;
		voices[voice].changes |= CH_PAN;
	}
}


static void AlsaVoicePlay(UBYTE voice,SWORD handle,ULONG start,ULONG size,
                            ULONG reppos,ULONG repend,UWORD flags)
{
	if ( start >= size)
		return;
	if ( flags & SF_LOOP ) {
		if ( repend > size ) 
			repend = size;
	}
			

	voices[ voice ].flags = flags;
	voices[ voice ].handle = handle;
	voices[ voice ].start = start;
	voices[ voice ].size = size;
	voices[ voice ].reppos = reppos;
	voices[ voice ].repend = repend;
	voices[ voice ].kick = 1;
}

/*
static void AlsaDummy()
{
	fflush(stdout);
}
*/


DRIVER drv_alsa={
	NULL,
	"ALSA sequencer driver",
	"ALSA Sequencer Driver v0.01 - by Jaroslav Kysela <perex@suse.cz>",
	AlsaIsThere,
	AlsaSampleLoad,
	AlsaSampleUnload,
	AlsaInit,
	AlsaExit,
	AlsaPlayStart,
	AlsaPlayStop,
	AlsaUpdate, 
	AlsaVoiceSetVolume,
	AlsaVoiceSetFrequency,
	AlsaVoiceSetPanning,
	AlsaVoicePlay
};
